# GDB 'explore' command.
# Copyright (C) 2012-2024 Free Software Foundation, Inc.

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Implementation of the GDB 'explore' command using the GDB Python API."""

import gdb


class Explorer(object):
    """Internal class which invokes other explorers."""

    # This map is filled by the Explorer.init_env() function
    type_code_to_explorer_map = {}

    _SCALAR_TYPE_LIST = (
        gdb.TYPE_CODE_CHAR,
        gdb.TYPE_CODE_INT,
        gdb.TYPE_CODE_BOOL,
        gdb.TYPE_CODE_FLT,
        gdb.TYPE_CODE_VOID,
        gdb.TYPE_CODE_ENUM,
    )

    @staticmethod
    def guard_expr(expr):
        length = len(expr)
        guard = False

        if expr[0] == "(" and expr[length - 1] == ")":
            pass
        else:
            i = 0
            while i < length:
                c = expr[i]
                if (
                    c == "_"
                    or ("a" <= c and c <= "z")
                    or ("A" <= c and c <= "Z")
                    or ("0" <= c and c <= "9")
                ):
                    pass
                else:
                    guard = True
                    break
                i += 1

        if guard:
            return "(" + expr + ")"
        else:
            return expr

    @staticmethod
    def explore_expr(expr, value, is_child):
        """Main function to explore an expression value.

        Arguments:
            expr: The expression string that is being explored.
            value: The gdb.Value value of the expression.
            is_child: Boolean value to indicate if the expression is a child.
                      An expression is a child if it is derived from the main
                      expression entered by the user.  For example, if the user
                      entered an expression which evaluates to a struct, then
                      when exploring the fields of the struct, is_child is set
                      to True internally.

        Returns:
            No return value.
        """
        type_code = value.type.code
        if type_code in Explorer.type_code_to_explorer_map:
            explorer_class = Explorer.type_code_to_explorer_map[type_code]
            while explorer_class.explore_expr(expr, value, is_child):
                pass
        else:
            print("Explorer for type '%s' not yet available.\n" % str(value.type))

    @staticmethod
    def explore_type(name, datatype, is_child):
        """Main function to explore a data type.

        Arguments:
            name: The string representing the path to the data type being
                  explored.
            datatype: The gdb.Type value of the data type being explored.
            is_child: Boolean value to indicate if the name is a child.
                      A name is a child if it is derived from the main name
                      entered by the user.  For example, if the user entered
                      the name of struct type, then when exploring the fields
                      of the struct, is_child is set to True internally.

        Returns:
            No return value.
        """
        type_code = datatype.code
        if type_code in Explorer.type_code_to_explorer_map:
            explorer_class = Explorer.type_code_to_explorer_map[type_code]
            while explorer_class.explore_type(name, datatype, is_child):
                pass
        else:
            print("Explorer for type '%s' not yet available.\n" % str(datatype))

    @staticmethod
    def init_env():
        """Initializes the Explorer environment.
        This function should be invoked before starting any exploration.  If
        invoked before an exploration, it need not be invoked for subsequent
        explorations.
        """
        Explorer.type_code_to_explorer_map = {
            gdb.TYPE_CODE_CHAR: ScalarExplorer,
            gdb.TYPE_CODE_INT: ScalarExplorer,
            gdb.TYPE_CODE_BOOL: ScalarExplorer,
            gdb.TYPE_CODE_FLT: ScalarExplorer,
            gdb.TYPE_CODE_VOID: ScalarExplorer,
            gdb.TYPE_CODE_ENUM: ScalarExplorer,
            gdb.TYPE_CODE_STRUCT: CompoundExplorer,
            gdb.TYPE_CODE_UNION: CompoundExplorer,
            gdb.TYPE_CODE_PTR: PointerExplorer,
            gdb.TYPE_CODE_REF: ReferenceExplorer,
            gdb.TYPE_CODE_RVALUE_REF: ReferenceExplorer,
            gdb.TYPE_CODE_TYPEDEF: TypedefExplorer,
            gdb.TYPE_CODE_ARRAY: ArrayExplorer,
        }

    @staticmethod
    def is_scalar_type(type):
        """Checks whether a type is a scalar type.
        A type is a scalar type of its type is
            gdb.TYPE_CODE_CHAR or
            gdb.TYPE_CODE_INT or
            gdb.TYPE_CODE_BOOL or
            gdb.TYPE_CODE_FLT or
            gdb.TYPE_CODE_VOID or
            gdb.TYPE_CODE_ENUM.

        Arguments:
            type: The type to be checked.

        Returns:
            'True' if 'type' is a scalar type. 'False' otherwise.
        """
        return type.code in Explorer._SCALAR_TYPE_LIST

    @staticmethod
    def return_to_parent_value():
        """A utility function which prints that the current exploration session
        is returning to the parent value. Useful when exploring values.
        """
        print("\nReturning to parent value...\n")

    @staticmethod
    def return_to_parent_value_prompt():
        """A utility function which prompts the user to press the 'enter' key
        so that the exploration session can shift back to the parent value.
        Useful when exploring values.
        """
        input("\nPress enter to return to parent value: ")

    @staticmethod
    def return_to_enclosing_type():
        """A utility function which prints that the current exploration session
        is returning to the enclosing type.  Useful when exploring types.
        """
        print("\nReturning to enclosing type...\n")

    @staticmethod
    def return_to_enclosing_type_prompt():
        """A utility function which prompts the user to press the 'enter' key
        so that the exploration session can shift back to the enclosing type.
        Useful when exploring types.
        """
        input("\nPress enter to return to enclosing type: ")


class ScalarExplorer(object):
    """Internal class used to explore scalar values."""

    @staticmethod
    def explore_expr(expr, value, is_child):
        """Function to explore scalar values.
        See Explorer.explore_expr and Explorer.is_scalar_type for more
        information.
        """
        print("'%s' is a scalar value of type '%s'." % (expr, value.type))
        print("%s = %s" % (expr, str(value)))

        if is_child:
            Explorer.return_to_parent_value_prompt()
            Explorer.return_to_parent_value()

        return False

    @staticmethod
    def explore_type(name, datatype, is_child):
        """Function to explore scalar types.
        See Explorer.explore_type and Explorer.is_scalar_type for more
        information.
        """
        if datatype.code == gdb.TYPE_CODE_ENUM:
            if is_child:
                print("%s is of an enumerated type '%s'." % (name, str(datatype)))
            else:
                print("'%s' is an enumerated type." % name)
        else:
            if is_child:
                print("%s is of a scalar type '%s'." % (name, str(datatype)))
            else:
                print("'%s' is a scalar type." % name)

        if is_child:
            Explorer.return_to_enclosing_type_prompt()
            Explorer.return_to_enclosing_type()

        return False


class PointerExplorer(object):
    """Internal class used to explore pointer values."""

    @staticmethod
    def explore_expr(expr, value, is_child):
        """Function to explore pointer values.
        See Explorer.explore_expr for more information.
        """
        print(
            "'%s' is a pointer to a value of type '%s'"
            % (expr, str(value.type.target()))
        )
        option = input(
            "Continue exploring it as a pointer to a single " "value [y/n]: "
        )
        if option == "y":
            deref_value = None
            try:
                deref_value = value.dereference()
                str(deref_value)
            except gdb.MemoryError:
                print(
                    "'%s' a pointer pointing to an invalid memory " "location." % expr
                )
                if is_child:
                    Explorer.return_to_parent_value_prompt()
                return False
            Explorer.explore_expr(
                "*%s" % Explorer.guard_expr(expr), deref_value, is_child
            )
            return False

        option = input("Continue exploring it as a pointer to an " "array [y/n]: ")
        if option == "y":
            while True:
                index = 0
                try:
                    index = int(
                        input(
                            "Enter the index of the element you "
                            "want to explore in '%s': " % expr
                        )
                    )
                except ValueError:
                    break
                element_expr = "%s[%d]" % (Explorer.guard_expr(expr), index)
                element = value[index]
                try:
                    str(element)
                except gdb.MemoryError:
                    print("Cannot read value at index %d." % index)
                    continue
                Explorer.explore_expr(element_expr, element, True)
            return False

        if is_child:
            Explorer.return_to_parent_value()
        return False

    @staticmethod
    def explore_type(name, datatype, is_child):
        """Function to explore pointer types.
        See Explorer.explore_type for more information.
        """
        target_type = datatype.target()
        print("\n%s is a pointer to a value of type '%s'." % (name, str(target_type)))

        Explorer.explore_type("the pointee type of %s" % name, target_type, is_child)
        return False


class ReferenceExplorer(object):
    """Internal class used to explore reference (TYPE_CODE_REF) values."""

    @staticmethod
    def explore_expr(expr, value, is_child):
        """Function to explore array values.
        See Explorer.explore_expr for more information.
        """
        referenced_value = value.referenced_value()
        Explorer.explore_expr(expr, referenced_value, is_child)
        return False

    @staticmethod
    def explore_type(name, datatype, is_child):
        """Function to explore pointer types.
        See Explorer.explore_type for more information.
        """
        target_type = datatype.target()
        Explorer.explore_type(name, target_type, is_child)
        return False


class ArrayExplorer(object):
    """Internal class used to explore arrays."""

    @staticmethod
    def explore_expr(expr, value, is_child):
        """Function to explore array values.
        See Explorer.explore_expr for more information.
        """
        target_type = value.type.target()
        print("'%s' is an array of '%s'." % (expr, str(target_type)))
        index = 0
        try:
            index = int(
                input(
                    "Enter the index of the element you want to "
                    "explore in '%s': " % expr
                )
            )
        except ValueError:
            if is_child:
                Explorer.return_to_parent_value()
            return False

        element = None
        try:
            element = value[index]
            str(element)
        except gdb.MemoryError:
            print("Cannot read value at index %d." % index)
            input("Press enter to continue... ")
            return True

        Explorer.explore_expr(
            "%s[%d]" % (Explorer.guard_expr(expr), index), element, True
        )
        return True

    @staticmethod
    def explore_type(name, datatype, is_child):
        """Function to explore array types.
        See Explorer.explore_type for more information.
        """
        target_type = datatype.target()
        print("%s is an array of '%s'." % (name, str(target_type)))

        Explorer.explore_type("the array element of %s" % name, target_type, is_child)
        return False


class CompoundExplorer(object):
    """Internal class used to explore struct, classes and unions."""

    @staticmethod
    def _print_fields(print_list):
        """Internal function which prints the fields of a struct/class/union."""
        max_field_name_length = 0
        for pair in print_list:
            if max_field_name_length < len(pair[0]):
                max_field_name_length = len(pair[0])

        for pair in print_list:
            print("  %*s = %s" % (max_field_name_length, pair[0], pair[1]))

    @staticmethod
    def _get_real_field_count(fields):
        real_field_count = 0
        for field in fields:
            if not field.artificial:
                real_field_count = real_field_count + 1

        return real_field_count

    @staticmethod
    def explore_expr(expr, value, is_child):
        """Function to explore structs/classes and union values.
        See Explorer.explore_expr for more information.
        """
        datatype = value.type
        type_code = datatype.code
        fields = datatype.fields()

        if type_code == gdb.TYPE_CODE_STRUCT:
            type_desc = "struct/class"
        else:
            type_desc = "union"

        if CompoundExplorer._get_real_field_count(fields) == 0:
            print(
                "The value of '%s' is a %s of type '%s' with no fields."
                % (expr, type_desc, str(value.type))
            )
            if is_child:
                Explorer.return_to_parent_value_prompt()
            return False

        print(
            "The value of '%s' is a %s of type '%s' with the following "
            "fields:\n" % (expr, type_desc, str(value.type))
        )

        has_explorable_fields = False
        choice_to_compound_field_map = {}
        current_choice = 0
        print_list = []
        for field in fields:
            if field.artificial:
                continue
            field_full_name = Explorer.guard_expr(expr) + "." + field.name
            if field.is_base_class:
                field_value = value.cast(field.type)
            else:
                field_value = value[field.name]
            literal_value = ""
            if type_code == gdb.TYPE_CODE_UNION:
                literal_value = "<Enter %d to explore this field of type " "'%s'>" % (
                    current_choice,
                    str(field.type),
                )
                has_explorable_fields = True
            else:
                if Explorer.is_scalar_type(field.type):
                    literal_value = "%s .. (Value of type '%s')" % (
                        str(field_value),
                        str(field.type),
                    )
                else:
                    if field.is_base_class:
                        field_desc = "base class"
                    else:
                        field_desc = "field"
                    literal_value = "<Enter %d to explore this %s of type " "'%s'>" % (
                        current_choice,
                        field_desc,
                        str(field.type),
                    )
                    has_explorable_fields = True

            choice_to_compound_field_map[str(current_choice)] = (
                field_full_name,
                field_value,
            )
            current_choice = current_choice + 1

            print_list.append((field.name, literal_value))

        CompoundExplorer._print_fields(print_list)
        print("")

        if has_explorable_fields:
            choice = input("Enter the field number of choice: ")
            if choice in choice_to_compound_field_map:
                Explorer.explore_expr(
                    choice_to_compound_field_map[choice][0],
                    choice_to_compound_field_map[choice][1],
                    True,
                )
                return True
            else:
                if is_child:
                    Explorer.return_to_parent_value()
        else:
            if is_child:
                Explorer.return_to_parent_value_prompt()

        return False

    @staticmethod
    def explore_type(name, datatype, is_child):
        """Function to explore struct/class and union types.
        See Explorer.explore_type for more information.
        """
        type_code = datatype.code
        type_desc = ""
        if type_code == gdb.TYPE_CODE_STRUCT:
            type_desc = "struct/class"
        else:
            type_desc = "union"

        fields = datatype.fields()
        if CompoundExplorer._get_real_field_count(fields) == 0:
            if is_child:
                print(
                    "%s is a %s of type '%s' with no fields."
                    % (name, type_desc, str(datatype))
                )
                Explorer.return_to_enclosing_type_prompt()
            else:
                print("'%s' is a %s with no fields." % (name, type_desc))
            return False

        if is_child:
            print(
                "%s is a %s of type '%s' "
                "with the following fields:\n" % (name, type_desc, str(datatype))
            )
        else:
            print("'%s' is a %s with the following " "fields:\n" % (name, type_desc))

        current_choice = 0
        choice_to_compound_field_map = {}
        print_list = []
        for field in fields:
            if field.artificial:
                continue
            if field.is_base_class:
                field_desc = "base class"
            else:
                field_desc = "field"
            rhs = "<Enter %d to explore this %s of type '%s'>" % (
                current_choice,
                field_desc,
                str(field.type),
            )
            print_list.append((field.name, rhs))
            choice_to_compound_field_map[str(current_choice)] = (
                field.name,
                field.type,
                field_desc,
            )
            current_choice = current_choice + 1

        CompoundExplorer._print_fields(print_list)
        print("")

        if len(choice_to_compound_field_map) > 0:
            choice = input("Enter the field number of choice: ")
            if choice in choice_to_compound_field_map:
                if is_child:
                    new_name = "%s '%s' of %s" % (
                        choice_to_compound_field_map[choice][2],
                        choice_to_compound_field_map[choice][0],
                        name,
                    )
                else:
                    new_name = "%s '%s' of '%s'" % (
                        choice_to_compound_field_map[choice][2],
                        choice_to_compound_field_map[choice][0],
                        name,
                    )
                Explorer.explore_type(
                    new_name, choice_to_compound_field_map[choice][1], True
                )
                return True
            else:
                if is_child:
                    Explorer.return_to_enclosing_type()
        else:
            if is_child:
                Explorer.return_to_enclosing_type_prompt()

        return False


class TypedefExplorer(object):
    """Internal class used to explore values whose type is a typedef."""

    @staticmethod
    def explore_expr(expr, value, is_child):
        """Function to explore typedef values.
        See Explorer.explore_expr for more information.
        """
        actual_type = value.type.strip_typedefs()
        print(
            "The value of '%s' is of type '%s' "
            "which is a typedef of type '%s'"
            % (expr, str(value.type), str(actual_type))
        )

        Explorer.explore_expr(expr, value.cast(actual_type), is_child)
        return False

    @staticmethod
    def explore_type(name, datatype, is_child):
        """Function to explore typedef types.
        See Explorer.explore_type for more information.
        """
        actual_type = datatype.strip_typedefs()
        if is_child:
            print(
                "The type of %s is a typedef of type '%s'." % (name, str(actual_type))
            )
        else:
            print("The type '%s' is a typedef of type '%s'." % (name, str(actual_type)))

        Explorer.explore_type(name, actual_type, is_child)
        return False


class ExploreUtils(object):
    """Internal class which provides utilities for the main command classes."""

    @staticmethod
    def check_args(name, arg_str):
        """Utility to check if adequate number of arguments are passed to an
        explore command.

        Arguments:
            name: The name of the explore command.
            arg_str: The argument string passed to the explore command.

        Returns:
            True if adequate arguments are passed, false otherwise.

        Raises:
            gdb.GdbError if adequate arguments are not passed.
        """
        if len(arg_str) < 1:
            raise gdb.GdbError("ERROR: '%s' requires an argument." % name)
            return False
        else:
            return True

    @staticmethod
    def get_type_from_str(type_str):
        """A utility function to deduce the gdb.Type value from a string
        representing the type.

        Arguments:
            type_str: The type string from which the gdb.Type value should be
                      deduced.

        Returns:
            The deduced gdb.Type value if possible, None otherwise.
        """
        try:
            # Assume the current language to be C/C++ and make a try.
            return gdb.parse_and_eval("(%s *)0" % type_str).type.target()
        except RuntimeError:
            # If assumption of current language to be C/C++ was wrong, then
            # lookup the type using the API.
            try:
                return gdb.lookup_type(type_str)
            except RuntimeError:
                return None

    @staticmethod
    def get_value_from_str(value_str):
        """A utility function to deduce the gdb.Value value from a string
        representing the value.

        Arguments:
            value_str: The value string from which the gdb.Value value should
                       be deduced.

        Returns:
            The deduced gdb.Value value if possible, None otherwise.
        """
        try:
            return gdb.parse_and_eval(value_str)
        except RuntimeError:
            return None


class ExploreCommand(gdb.Command):
    """Explore a value or a type valid in the current context.

    Usage: explore ARG

    - ARG is either a valid expression or a type name.
    - At any stage of exploration, hit the return key (instead of a
    choice, if any) to return to the enclosing type or value."""

    def __init__(self):
        super(ExploreCommand, self).__init__(
            name="explore", command_class=gdb.COMMAND_DATA, prefix=True
        )

    def invoke(self, arg_str, from_tty):
        if ExploreUtils.check_args("explore", arg_str) is False:
            return

        # Check if it is a value
        value = ExploreUtils.get_value_from_str(arg_str)
        if value is not None:
            Explorer.explore_expr(arg_str, value, False)
            return

        # If it is not a value, check if it is a type
        datatype = ExploreUtils.get_type_from_str(arg_str)
        if datatype is not None:
            Explorer.explore_type(arg_str, datatype, False)
            return

        # If it is neither a value nor a type, raise an error.
        raise gdb.GdbError(
            (
                "'%s' neither evaluates to a value nor is a type "
                "in the current context." % arg_str
            )
        )


class ExploreValueCommand(gdb.Command):
    """Explore value of an expression valid in the current context.

    Usage: explore value ARG

    - ARG is a valid expression.
    - At any stage of exploration, hit the return key (instead of a
    choice, if any) to return to the enclosing value."""

    def __init__(self):
        super(ExploreValueCommand, self).__init__(
            name="explore value", command_class=gdb.COMMAND_DATA
        )

    def invoke(self, arg_str, from_tty):
        if ExploreUtils.check_args("explore value", arg_str) is False:
            return

        value = ExploreUtils.get_value_from_str(arg_str)
        if value is None:
            raise gdb.GdbError(
                (
                    " '%s' does not evaluate to a value in the current "
                    "context." % arg_str
                )
            )
            return

        Explorer.explore_expr(arg_str, value, False)


class ExploreTypeCommand(gdb.Command):
    """Explore a type or the type of an expression.

    Usage: explore type ARG

    - ARG is a valid expression or a type name.
    - At any stage of exploration, hit the return key (instead of a
    choice, if any) to return to the enclosing type."""

    def __init__(self):
        super(ExploreTypeCommand, self).__init__(
            name="explore type", command_class=gdb.COMMAND_DATA
        )

    def invoke(self, arg_str, from_tty):
        if ExploreUtils.check_args("explore type", arg_str) is False:
            return

        datatype = ExploreUtils.get_type_from_str(arg_str)
        if datatype is not None:
            Explorer.explore_type(arg_str, datatype, False)
            return

        value = ExploreUtils.get_value_from_str(arg_str)
        if value is not None:
            print("'%s' is of type '%s'." % (arg_str, str(value.type)))
            Explorer.explore_type(str(value.type), value.type, False)
            return

        raise gdb.GdbError(
            ("'%s' is not a type or value in the current " "context." % arg_str)
        )


Explorer.init_env()

ExploreCommand()
ExploreValueCommand()
ExploreTypeCommand()



# Copyright 2023-2024 Free Software Foundation, Inc.

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import gdb

from .server import capability, request
from .startup import in_gdb_thread


@in_gdb_thread
def module_id(objfile):
    """Return the module ID for the objfile."""
    return objfile.username


@in_gdb_thread
def is_module(objfile):
    """Return True if OBJFILE represents a valid Module."""
    return objfile.is_valid() and objfile.owner is None


@in_gdb_thread
def make_module(objf):
    """Return a Module representing the objfile OBJF.

    The objfile must pass the 'is_module' test."""
    result = {
        "id": module_id(objf),
        "name": objf.username,
    }
    if objf.is_file:
        result["path"] = objf.filename
    return result


@capability("supportsModulesRequest")
@request("modules")
def modules(*, startModule: int = 0, moduleCount: int = 0, **args):
    # Don't count invalid objfiles or separate debug objfiles.
    objfiles = [x for x in gdb.objfiles() if is_module(x)]
    if moduleCount == 0:
        # Use all items.
        last = len(objfiles)
    else:
        last = startModule + moduleCount
    return {
        "modules": [make_module(x) for x in objfiles[startModule:last]],
        "totalModules": len(objfiles),
    }



# Copyright 2022-2024 Free Software Foundation, Inc.

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import functools
import heapq
import inspect
import json
import threading
from contextlib import contextmanager

import gdb

from .io import read_json, start_json_writer
from .startup import (
    DAPException,
    DAPQueue,
    LogLevel,
    exec_and_log,
    in_dap_thread,
    in_gdb_thread,
    log,
    log_stack,
    start_thread,
    thread_log,
)
from .typecheck import type_check

# Map capability names to values.
_capabilities = {}

# Map command names to callables.
_commands = {}

# The global server.
_server = None

# This is set by the initialize request and is used when rewriting
# line numbers.
_lines_start_at_1 = False


class DeferredRequest:
    """If a DAP request function returns a deferred request, no
    response is sent immediately.

    Instead, request processing continues, with this particular
    request remaining un-replied-to.

    Later, when the result is available, the deferred request can be
    scheduled.  This causes 'invoke' to be called and then the
    response to be sent to the client.

    """

    # This is for internal use by the server.  It should not be
    # overridden by any subclass.  This adds the request ID and the
    # result template object to this object.  These are then used
    # during rescheduling.
    def set_request(self, req, result):
        self._req = req
        self._result = result

    @in_dap_thread
    def invoke(self):
        """Implement the deferred request.

        This will be called from 'reschedule' (and should not be
        called elsewhere).  It should return the 'body' that will be
        sent in the response.  None means no 'body' field will be set.

        Subclasses must override this.

        """
        pass

    @in_dap_thread
    def reschedule(self):
        """Call this to reschedule this deferred request.

        This will call 'invoke' after the appropriate bookkeeping and
        will arrange for its result to be reported to the client.

        """
        with _server.canceller.current_request(self._req):
            _server.invoke_request(self._req, self._result, self.invoke)


# A subclass of Exception that is used solely for reporting that a
# request needs the inferior to be stopped, but it is not stopped.
class NotStoppedException(Exception):
    pass


# This is used to handle cancellation requests.  It tracks all the
# needed state, so that we can cancel both requests that are in flight
# as well as queued requests.
class CancellationHandler:
    def __init__(self):
        # Methods on this class acquire this lock before proceeding.
        # A recursive lock is used to simplify the 'check_cancel'
        # callers.
        self.lock = threading.RLock()
        # The request currently being handled, or None.
        self.in_flight_dap_thread = None
        self.in_flight_gdb_thread = None
        self.reqs = []
        # A set holding the request IDs of all deferred requests that
        # are still unresolved.
        self.deferred_ids = set()

    @contextmanager
    def current_request(self, req):
        """Return a new context manager that registers that request
        REQ has started."""
        try:
            with self.lock:
                self.in_flight_dap_thread = req
            # Note we do not call check_cancel here.  This is a bit of
            # a hack, but it's because the direct callers of this
            # aren't prepared for a KeyboardInterrupt.
            yield
        finally:
            with self.lock:
                self.in_flight_dap_thread = None

    def defer_request(self, req):
        """Indicate that the request REQ has been deferred."""
        with self.lock:
            self.deferred_ids.add(req)

    def request_finished(self, req):
        """Indicate that the request REQ is finished.

        It doesn't matter whether REQ succeeded or failed, only that
        processing for it is done.

        """
        with self.lock:
            # Use discard here, not remove, because this is called
            # regardless of whether REQ was deferred.
            self.deferred_ids.discard(req)

    def check_cancel(self, req):
        """Check whether request REQ is cancelled.
        If so, raise KeyboardInterrupt."""
        with self.lock:
            # We want to drop any cancellations that come before REQ,
            # but keep ones for any deferred requests that are still
            # unresolved.  This holds any such requests that were
            # popped during the loop.
            deferred = []
            try:
                # If the request is cancelled, don't execute the region.
                while len(self.reqs) > 0 and self.reqs[0] <= req:
                    # In most cases, if we see a cancellation request
                    # on the heap that is before REQ, we can just
                    # ignore it -- we missed our chance to cancel that
                    # request.
                    next_id = heapq.heappop(self.reqs)
                    if next_id == req:
                        raise KeyboardInterrupt()
                    elif next_id in self.deferred_ids:
                        # We could be in a situation where we're
                        # processing request 23, but request 18 is
                        # still deferred.  In this case, popping
                        # request 18 here will lose the cancellation.
                        # So, we preserve it.
                        deferred.append(next_id)
            finally:
                for x in deferred:
                    heapq.heappush(self.reqs, x)

    def cancel(self, req):
        """Call to cancel a request.

        If the request has already finished, this is ignored.
        If the request is in flight, it is interrupted.
        If the request has not yet been seen, the cancellation is queued."""
        with self.lock:
            if req == self.in_flight_gdb_thread:
                gdb.interrupt()
            else:
                # We don't actually ignore the request here, but in
                # the 'check_cancel' method.  This way we don't have to
                # track as much state.  Also, this implementation has
                # the weird property that a request can be cancelled
                # before it is even sent.  It didn't seem worthwhile
                # to try to check for this.
                heapq.heappush(self.reqs, req)

    @contextmanager
    def interruptable_region(self, req):
        """Return a new context manager that sets in_flight_gdb_thread to
        REQ."""
        if req is None:
            # No request is handled in the region, just execute the region.
            yield
            return
        try:
            with self.lock:
                self.check_cancel(req)
                # Request is being handled by the gdb thread.
                self.in_flight_gdb_thread = req
            # Execute region.  This may be interrupted by gdb.interrupt.
            yield
        finally:
            with self.lock:
                # Request is no longer handled by the gdb thread.
                self.in_flight_gdb_thread = None


class Server:
    """The DAP server class."""

    def __init__(self, in_stream, out_stream, child_stream):
        self.in_stream = in_stream
        self.out_stream = out_stream
        self.child_stream = child_stream
        self.delayed_fns_lock = threading.Lock()
        self.defer_stop_events = False
        self.delayed_fns = []
        # This queue accepts JSON objects that are then sent to the
        # DAP client.  Writing is done in a separate thread to avoid
        # blocking the read loop.
        self.write_queue = DAPQueue()
        # Reading is also done in a separate thread, and a queue of
        # requests is kept.
        self.read_queue = DAPQueue()
        self.done = False
        self.canceller = CancellationHandler()
        global _server
        _server = self

    # A helper for request processing.  REQ is the request ID.  RESULT
    # is a result "template" -- a dictionary with a few items already
    # filled in.  This helper calls FN and then fills in the remaining
    # parts of RESULT, as needed.  If FN returns an ordinary result,
    # or if it fails, then the final RESULT is sent as a response to
    # the client.  However, if FN returns a DeferredRequest, then that
    # request is updated (see DeferredRequest.set_request) and no
    # response is sent.
    @in_dap_thread
    def invoke_request(self, req, result, fn):
        try:
            self.canceller.check_cancel(req)
            fn_result = fn()
            result["success"] = True
            if isinstance(fn_result, DeferredRequest):
                fn_result.set_request(req, result)
                self.canceller.defer_request(req)
                # Do not send a response.
                return
            elif fn_result is not None:
                result["body"] = fn_result
        except NotStoppedException:
            # This is an expected exception, and the result is clearly
            # visible in the log, so do not log it.
            result["success"] = False
            result["message"] = "notStopped"
        except KeyboardInterrupt:
            # This can only happen when a request has been canceled.
            result["success"] = False
            result["message"] = "cancelled"
        except DAPException as e:
            # Don't normally want to see this, as it interferes with
            # the test suite.
            log_stack(LogLevel.FULL)
            result["success"] = False
            result["message"] = str(e)
        except BaseException as e:
            log_stack()
            result["success"] = False
            result["message"] = str(e)

        self.canceller.request_finished(req)
        # We have a response for the request, so send it back to the
        # client.
        self._send_json(result)

    # Treat PARAMS as a JSON-RPC request and perform its action.
    # PARAMS is just a dictionary from the JSON.
    @in_dap_thread
    def _handle_command(self, params):
        req = params["seq"]
        result = {
            "request_seq": req,
            "type": "response",
            "command": params["command"],
        }

        if "arguments" in params:
            args = params["arguments"]
        else:
            args = {}

        def fn():
            global _commands
            return _commands[params["command"]](**args)

        self.invoke_request(req, result, fn)

    # Read inferior output and sends OutputEvents to the client.  It
    # is run in its own thread.
    def _read_inferior_output(self):
        while True:
            line = self.child_stream.readline()
            self.send_event(
                "output",
                {
                    "category": "stdout",
                    "output": line,
                },
            )

    # Send OBJ to the client, logging first if needed.
    def _send_json(self, obj):
        log("WROTE: <<<" + json.dumps(obj) + ">>>")
        self.write_queue.put(obj)

    # This is run in a separate thread and simply reads requests from
    # the client and puts them into a queue.  A separate thread is
    # used so that 'cancel' requests can be handled -- the DAP thread
    # will normally block, waiting for each request to complete.
    def _reader_thread(self):
        while True:
            cmd = read_json(self.in_stream)
            if cmd is None:
                break
            log("READ: <<<" + json.dumps(cmd) + ">>>")
            # Be extra paranoid about the form here.  If anything is
            # missing, it will be put in the queue and then an error
            # issued by ordinary request processing.
            if (
                "command" in cmd
                and cmd["command"] == "cancel"
                and "arguments" in cmd
                # gdb does not implement progress, so there's no need
                # to check for progressId.
                and "requestId" in cmd["arguments"]
            ):
                self.canceller.cancel(cmd["arguments"]["requestId"])
            self.read_queue.put(cmd)
        # When we hit EOF, signal it with None.
        self.read_queue.put(None)

    @in_dap_thread
    def main_loop(self):
        """The main loop of the DAP server."""
        # Before looping, start the thread that writes JSON to the
        # client, and the thread that reads output from the inferior.
        start_thread("output reader", self._read_inferior_output)
        json_writer = start_json_writer(self.out_stream, self.write_queue)
        start_thread("JSON reader", self._reader_thread)
        while not self.done:
            cmd = self.read_queue.get()
            # A None value here means the reader hit EOF.
            if cmd is None:
                break
            req = cmd["seq"]
            with self.canceller.current_request(req):
                self._handle_command(cmd)
            fns = None
            with self.delayed_fns_lock:
                fns = self.delayed_fns
                self.delayed_fns = []
                self.defer_stop_events = False
            for fn in fns:
                fn()
        # Got the terminate request.  This is handled by the
        # JSON-writing thread, so that we can ensure that all
        # responses are flushed to the client before exiting.
        self.write_queue.put(None)
        json_writer.join()
        send_gdb("quit")

    @in_dap_thread
    def send_event_later(self, event, body=None):
        """Send a DAP event back to the client, but only after the
        current request has completed."""
        with self.delayed_fns_lock:
            self.delayed_fns.append(lambda: self.send_event(event, body))

    @in_gdb_thread
    def send_event_maybe_later(self, event, body=None):
        """Send a DAP event back to the client, but if a request is in-flight
        within the dap thread and that request is configured to delay the event,
        wait until the response has been sent until the event is sent back to
        the client."""
        with self.canceller.lock:
            if self.canceller.in_flight_dap_thread:
                with self.delayed_fns_lock:
                    if self.defer_stop_events:
                        self.delayed_fns.append(lambda: self.send_event(event, body))
                        return
        self.send_event(event, body)

    @in_dap_thread
    def call_function_later(self, fn):
        """Call FN later -- after the current request's response has been sent."""
        with self.delayed_fns_lock:
            self.delayed_fns.append(fn)

    # Note that this does not need to be run in any particular thread,
    # because it just creates an object and writes it to a thread-safe
    # queue.
    def send_event(self, event, body=None):
        """Send an event to the DAP client.
        EVENT is the name of the event, a string.
        BODY is the body of the event, an arbitrary object."""
        obj = {
            "type": "event",
            "event": event,
        }
        if body is not None:
            obj["body"] = body
        self._send_json(obj)

    def shutdown(self):
        """Request that the server shut down."""
        # Just set a flag.  This operation is complicated because we
        # want to write the result of the request before exiting.  See
        # main_loop.
        self.done = True


def send_event(event, body=None):
    """Send an event to the DAP client.
    EVENT is the name of the event, a string.
    BODY is the body of the event, an arbitrary object."""
    global _server
    _server.send_event(event, body)


def send_event_maybe_later(event, body=None):
    """Send a DAP event back to the client, but if a request is in-flight
    within the dap thread and that request is configured to delay the event,
    wait until the response has been sent until the event is sent back to
    the client."""
    global _server
    _server.send_event_maybe_later(event, body)


def call_function_later(fn):
    """Call FN later -- after the current request's response has been sent."""
    global _server
    _server.call_function_later(fn)


# A helper decorator that checks whether the inferior is running.
def _check_not_running(func):
    @functools.wraps(func)
    def check(*args, **kwargs):
        # Import this as late as possible.  This is done to avoid
        # circular imports.
        from .events import inferior_running

        if inferior_running:
            raise NotStoppedException()
        return func(*args, **kwargs)

    return check


def request(
    name: str,
    *,
    response: bool = True,
    on_dap_thread: bool = False,
    expect_stopped: bool = True,
    defer_stop_events: bool = False
):
    """A decorator for DAP requests.

    This registers the function as the implementation of the DAP
    request NAME.  By default, the function is invoked in the gdb
    thread, and its result is returned as the 'body' of the DAP
    response.

    Some keyword arguments are provided as well:

    If RESPONSE is False, the result of the function will not be
    waited for and no 'body' will be in the response.

    If ON_DAP_THREAD is True, the function will be invoked in the DAP
    thread.  When ON_DAP_THREAD is True, RESPONSE may not be False.

    If EXPECT_STOPPED is True (the default), then the request will
    fail with the 'notStopped' reason if it is processed while the
    inferior is running.  When EXPECT_STOPPED is False, the request
    will proceed regardless of the inferior's state.

    If DEFER_STOP_EVENTS is True, then make sure any stop events sent
    during the request processing are not sent to the client until the
    response has been sent.
    """

    # Validate the parameters.
    assert not on_dap_thread or response

    def wrap(func):
        code = func.__code__
        # We don't permit requests to have positional arguments.
        try:
            assert code.co_posonlyargcount == 0
        except AttributeError:
            # Attribute co_posonlyargcount is supported starting python 3.8.
            pass
        assert code.co_argcount == 0
        # A request must have a **args parameter.
        assert code.co_flags & inspect.CO_VARKEYWORDS

        # Type-check the calls.
        func = type_check(func)

        # Verify that the function is run on the correct thread.
        if on_dap_thread:
            cmd = in_dap_thread(func)
        else:
            func = in_gdb_thread(func)

            if response:
                if defer_stop_events:
                    global _server
                    if _server is not None:
                        with _server.delayed_events_lock:
                            _server.defer_stop_events = True

                def sync_call(**args):
                    return send_gdb_with_response(lambda: func(**args))

                cmd = sync_call
            else:

                def non_sync_call(**args):
                    return send_gdb(lambda: func(**args))

                cmd = non_sync_call

        # If needed, check that the inferior is not running.  This
        # wrapping is done last, so the check is done first, before
        # trying to dispatch the request to another thread.
        if expect_stopped:
            cmd = _check_not_running(cmd)

        global _commands
        assert name not in _commands
        _commands[name] = cmd
        return cmd

    return wrap


def capability(name, value=True):
    """A decorator that indicates that the wrapper function implements
    the DAP capability NAME."""

    def wrap(func):
        global _capabilities
        assert name not in _capabilities
        _capabilities[name] = value
        return func

    return wrap


def client_bool_capability(name, default=False):
    """Return the value of a boolean client capability.

    If the capability was not specified, or did not have boolean type,
    DEFAULT is returned.  DEFAULT defaults to False."""
    global _server
    if name in _server.config and isinstance(_server.config[name], bool):
        return _server.config[name]
    return default


@request("initialize", on_dap_thread=True)
def initialize(**args):
    global _server, _capabilities
    _server.config = args
    _server.send_event_later("initialized")
    global _lines_start_at_1
    _lines_start_at_1 = client_bool_capability("linesStartAt1", True)
    return _capabilities.copy()


@request("terminate", expect_stopped=False)
@capability("supportsTerminateRequest")
def terminate(**args):
    exec_and_log("kill")


@request("disconnect", on_dap_thread=True, expect_stopped=False)
@capability("supportTerminateDebuggee")
def disconnect(*, terminateDebuggee: bool = False, **args):
    if terminateDebuggee:
        send_gdb_with_response("kill")
    _server.shutdown()


@request("cancel", on_dap_thread=True, expect_stopped=False)
@capability("supportsCancelRequest")
def cancel(**args):
    # If a 'cancel' request can actually be satisfied, it will be
    # handled specially in the reader thread.  However, in order to
    # construct a proper response, the request is also added to the
    # command queue and so ends up here.  Additionally, the spec says:
    #    The cancel request may return an error if it could not cancel
    #    an operation but a client should refrain from presenting this
    #    error to end users.
    # ... which gdb takes to mean that it is fine for all cancel
    # requests to report success.
    return None


class Invoker(object):
    """A simple class that can invoke a gdb command."""

    def __init__(self, cmd):
        self.cmd = cmd

    # This is invoked in the gdb thread to run the command.
    @in_gdb_thread
    def __call__(self):
        exec_and_log(self.cmd)


class Cancellable(object):

    def __init__(self, fn, result_q=None):
        self.fn = fn
        self.result_q = result_q
        with _server.canceller.lock:
            self.req = _server.canceller.in_flight_dap_thread

    # This is invoked in the gdb thread to run self.fn.
    @in_gdb_thread
    def __call__(self):
        try:
            with _server.canceller.interruptable_region(self.req):
                val = self.fn()
                if self.result_q is not None:
                    self.result_q.put(val)
        except (Exception, KeyboardInterrupt) as e:
            if self.result_q is not None:
                # Pass result or exception to caller.
                self.result_q.put(e)
            elif isinstance(e, KeyboardInterrupt):
                # Fn was cancelled.
                pass
            else:
                # Exception happened.  Ignore and log it.
                err_string = "%s, %s" % (e, type(e))
                thread_log("caught exception: " + err_string)
                log_stack()


def send_gdb(cmd):
    """Send CMD to the gdb thread.
    CMD can be either a function or a string.
    If it is a string, it is passed to gdb.execute."""
    if isinstance(cmd, str):
        cmd = Invoker(cmd)

    # Post the event and don't wait for the result.
    gdb.post_event(Cancellable(cmd))


def send_gdb_with_response(fn):
    """Send FN to the gdb thread and return its result.
    If FN is a string, it is passed to gdb.execute and None is
    returned as the result.
    If FN throws an exception, this function will throw the
    same exception in the calling thread.
    """
    if isinstance(fn, str):
        fn = Invoker(fn)

    # Post the event and wait for the result in result_q.
    result_q = DAPQueue()
    gdb.post_event(Cancellable(fn, result_q))
    val = result_q.get()

    if isinstance(val, (Exception, KeyboardInterrupt)):
        raise val
    return val


def export_line(line):
    """Rewrite LINE according to client capability.
    This applies the linesStartAt1 capability as needed,
    when sending a line number from gdb to the client."""
    global _lines_start_at_1
    if not _lines_start_at_1:
        # In gdb, lines start at 1, so we only need to change this if
        # the client starts at 0.
        line = line - 1
    return line


def import_line(line):
    """Rewrite LINE according to client capability.
    This applies the linesStartAt1 capability as needed,
    when the client sends a line number to gdb."""
    global _lines_start_at_1
    if not _lines_start_at_1:
        # In gdb, lines start at 1, so we only need to change this if
        # the client starts at 0.
        line = line + 1
    return line



# ROCm OCP microscaling floating point convenience functions.
# Copyright (C) 2025 Free Software Foundation, Inc.

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""$_mx_get_scaled_float"""

import math
import re

import gdb


def extract_mx_bits(val, sgn_msk, exp_msk, mnts_msk):
    """
    Extract sign, exponent, and mantissa bits from a given value.

    Arguments:
    val      -- Input 8-bit value from which bits are extracted.
    sgn_msk  -- Bitmask to isolate the sign bit(s).
    exp_msk  -- Bitmask to isolate the exponent bits.
    mnts_msk -- Bitmask to isolate the mantissa bits.

    Returns:
    A tuple (sgn, exp, mnts) containing the extracted sign, exponent, and mantissa.
    """
    val &= 0xFF
    ctz = (sgn_msk & -sgn_msk).bit_length() - 1
    sgn = (val & sgn_msk) >> ctz
    ctz = (exp_msk & -exp_msk).bit_length() - 1
    exp = (val & exp_msk) >> ctz
    ctz = (mnts_msk & -mnts_msk).bit_length() - 1
    mnts = (val & mnts_msk) >> ctz
    return sgn, exp, mnts


def compute_mx_value(sign, exponent, bias, mantissa, mbits):
    """Calculate microscaling subnormal or normal value"""
    if exponent == 0:
        result = ((-1) ** sign) * (2 ** (1 - bias)) * ((2**-mbits) * mantissa)
    else:
        result = (
            ((-1) ** sign)
            * (2 ** (exponent - bias))
            * (1 + 2 ** (-1 * mbits) * mantissa)
        )
    return result


def compute_fp8e4m3(val):
    bias = 7
    mbits = 3
    sign, exponent, mantissa = extract_mx_bits(val, 0x80, 0x78, 0x07)
    if exponent == 0xF and mantissa == 0x7:
        return math.nan
    if exponent == 0 and mantissa == 0:
        return float(0)
    return compute_mx_value(sign, exponent, bias, mantissa, mbits)


def mx_fp8e4m3(val):
    fp = [compute_fp8e4m3(int(val))]
    return fp


def mx_fp8x2e4m3(val):
    fp = [compute_fp8e4m3(int(val) >> (8 if k else 0)) for k in range(0, 2)]
    return fp


def compute_fp8e5m2(val):
    bias = 15
    mbits = 2
    sign, exponent, mantissa = extract_mx_bits(val, 0x80, 0x7C, 0x03)
    if exponent == 0x1F and mantissa == 0:
        return -math.inf if sign else math.inf
    if exponent == 0x1F and (mantissa == 1 or mantissa == 2 or mantissa == 3):
        return math.nan
    if exponent == 0 and mantissa == 0:
        return float(0)
    return compute_mx_value(sign, exponent, bias, mantissa, mbits)


def mx_fp8e5m2(val):
    fp = [compute_fp8e5m2(int(val))]
    return fp


def mx_fp8x2e5m2(val):
    fp = [compute_fp8e5m2(int(val) >> (8 if k else 0)) for k in range(0, 2)]
    return fp


def compute_fp6e2m3(val):
    bias = 1
    mbits = 3
    sign, exponent, mantissa = extract_mx_bits(val, 0x20, 0x18, 0x07)
    if exponent == 0 and mantissa == 0:
        return float(0)
    return compute_mx_value(sign, exponent, bias, mantissa, mbits)


def mx_fp6x32e2m3(val):
    v192bits = (
        (int(val[5]) << 160)
        | (int(val[4]) << 128)
        | (int(val[3]) << 96)
        | (int(val[2]) << 64)
        | (int(val[1]) << 32)
        | (int(val[0]) << 0)
    )
    fp = [compute_fp6e2m3((v192bits >> (6 * k)) & 0xFF) for k in range(0, 32)]
    return fp


def compute_fp6e3m2(val):
    bias = 3
    mbits = 2
    sign, exponent, mantissa = extract_mx_bits(val, 0x20, 0x1C, 0x03)
    if exponent == 0 and mantissa == 0:
        return float(0)
    return compute_mx_value(sign, exponent, bias, mantissa, mbits)


def mx_fp6x32e3m2(val):
    v192bits = (
        (int(val[5]) << 160)
        | (int(val[4]) << 128)
        | (int(val[3]) << 96)
        | (int(val[2]) << 64)
        | (int(val[1]) << 32)
        | (int(val[0]) << 0)
    )
    fp = [compute_fp6e3m2((v192bits >> (6 * k)) & 0xFF) for k in range(0, 32)]
    return fp


def compute_fp4e2m1(val):
    bias = 1
    mbits = 1
    sign, exponent, mantissa = extract_mx_bits(val, 0x8, 0x6, 0x1)
    if exponent == 0 and mantissa == 0:
        return float(0)
    else:
        return compute_mx_value(sign, exponent, bias, mantissa, mbits)


def mx_fp4x2e2m1(val):
    fp = [compute_fp4e2m1(int(val) >> (4 if k else 0)) for k in range(0, 2)]
    return fp


def compute_uint8e8m0(val):
    v = int(val)
    if v < -127 or v >= 128:
        return math.nan
    else:
        return float(2 ** (float(v) + 127))


def extract_mx_values(mxvar):
    if re.match(r"^__hipext_ocp_fp8_e4m3$", str(mxvar.type)):
        return mx_fp8e4m3(mxvar["__x"])

    if re.match(r"^__hipext_ocp_fp8x2_e4m3$", str(mxvar.type)):
        return mx_fp8x2e4m3(mxvar["__x"])

    if re.match(r"^__hipext_ocp_fp8_e5m2$", str(mxvar.type)):
        return mx_fp8e5m2(mxvar["__x"])

    if re.match(r"^__hipext_ocp_fp8x2_e5m2$", str(mxvar.type)):
        return mx_fp8x2e5m2(mxvar["__x"])

    if re.match(r"^__hipext_ocp_fp6x32_e2m3$", str(mxvar.type)):
        return mx_fp6x32e2m3(mxvar["__x"])

    if re.match(r"^__hipext_ocp_fp6x32_e3m2$", str(mxvar.type)):
        return mx_fp6x32e3m2(mxvar["__x"])

    if re.match(r"^__hipext_ocp_fp4x2_e2m1$", str(mxvar.type)):
        return mx_fp4x2e2m1(mxvar["__x"])

    raise TypeError("unknown ROCm microscaling scalar type")


def hipexp_ocp_fp_re():
    return r"^__hipext_ocp_fp\d+(x\d+)?_e\dm\d$"


class _MxGetScaledFloat(gdb.Function):
    """$_mx_get_scaled_float - get scaled values(s) of ROCm microscaling objects.

    Usage: $_mx_get_scaled_float(mx_scalar, mx_scale=1)

    Returns:
      Scaled value(s) of the ROCm microscaling scalar object"""

    def __init__(self):
        super().__init__("_mx_get_scaled_float")

    def invoke(self, mxvar, mxscale=None):
        if not re.match(hipexp_ocp_fp_re(), str(mxvar.type)):
            raise TypeError("mx_scalar must be of ROCm microscaling scalar type")

        if mxscale is None:
            mxscale = gdb.Value(1.0).cast(gdb.lookup_type("float"))
        mxscalev = None

        # If mxscale is __amd_scale_t var, convert its bit encoding into float value.
        # Otherwise, mxscale defaults to 1 or is a user specified int/float value.
        if re.match(r"^__amd_scale_t$", str(mxscale.type)):
            mxscalev = compute_uint8e8m0(mxscale)
        else:
            mxscalev = float(mxscale)
        if not mxscalev >= 1:
            raise ValueError("usage: _mx_get_scaled_float(mx_scalar, mx_scale >= 1)")

        # Extract the mx var values into list, create list-sized bytearray.
        mxvarvl = extract_mx_values(mxvar)
        ftype = gdb.lookup_type("float")
        vtype = ftype.vector(len(mxvarvl) - 1)
        barray = bytearray(vtype.sizeof)

        # Convert values list to gdb.Value list, then copy it into a bytearray.
        # Below float(v) is a host (python) float type, created from the lower
        # precision inferior microscaling type.
        gvals = [gdb.Value(float(v) * mxscalev).cast(ftype) for v in mxvarvl]
        fsize = ftype.sizeof
        for i, val in enumerate(gvals):
            barray[i * fsize : (i + 1) * fsize] = gvals[i].bytes

        # Convert the bytearray into the returned gdb.Value.
        return gdb.Value(barray, vtype)


# GDB will import us automagically via gdb/__init__.py.
_MxGetScaledFloat()



# Extended prompt.
# Copyright (C) 2011-2024 Free Software Foundation, Inc.

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""GDB command for working with extended prompts."""

import gdb
import gdb.prompt


class _ExtendedPrompt(gdb.Parameter):
    """Set the extended prompt.

    Usage: set extended-prompt VALUE

    Substitutions are applied to VALUE to compute the real prompt.

    The currently defined substitutions are:"""

    # Add the prompt library's dynamically generated help to the
    # __doc__ string.
    __doc__ = __doc__ + "\n" + gdb.prompt.prompt_help()

    set_doc = "Set the extended prompt."
    show_doc = "Show the extended prompt."

    def __init__(self):
        super(_ExtendedPrompt, self).__init__(
            "extended-prompt", gdb.COMMAND_SUPPORT, gdb.PARAM_STRING_NOESCAPE
        )
        self.value = ""
        self.hook_set = False

    def get_show_string(self, pvalue):
        if self.value:
            return "The extended prompt is: " + self.value
        else:
            return "The extended prompt is not set."

    def get_set_string(self):
        if self.hook_set is False:
            gdb.prompt_hook = self.before_prompt_hook
            self.hook_set = True
        return ""

    def before_prompt_hook(self, current):
        if self.value:
            return gdb.prompt.substitute_prompt(self.value)
        else:
            return None


_ExtendedPrompt()



# Copyright 2022-2024 Free Software Foundation, Inc.

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import gdb

from .frames import frame_for_id
from .globalvars import get_global_scope
from .server import export_line, request
from .sources import make_source
from .startup import in_gdb_thread
from .varref import BaseReference

# Map DAP frame IDs to scopes.  This ensures that scopes are re-used.
frame_to_scope = {}


# If the most recent stop was due to a 'finish', and the return value
# could be computed, then this holds that value.  Otherwise it holds
# None.
_last_return_value = None


# When the inferior is re-started, we erase all scope references.  See
# the section "Lifetime of Objects References" in the spec.
@in_gdb_thread
def clear_scopes(event):
    global frame_to_scope
    frame_to_scope = {}
    global _last_return_value
    _last_return_value = None


gdb.events.cont.connect(clear_scopes)


@in_gdb_thread
def set_finish_value(val):
    """Set the current 'finish' value on a stop."""
    global _last_return_value
    _last_return_value = val


# A helper function to compute the value of a symbol.  SYM is either a
# gdb.Symbol, or an object implementing the SymValueWrapper interface.
# FRAME is a frame wrapper, as produced by a frame filter.  Returns a
# tuple of the form (NAME, VALUE), where NAME is the symbol's name and
# VALUE is a gdb.Value.
@in_gdb_thread
def symbol_value(sym, frame):
    inf_frame = frame.inferior_frame()
    # Make sure to select the frame first.  Ideally this would not
    # be needed, but this is a way to set the current language
    # properly so that language-dependent APIs will work.
    inf_frame.select()
    name = str(sym.symbol())
    val = sym.value()
    if val is None:
        # No synthetic value, so must read the symbol value
        # ourselves.
        val = sym.symbol().value(inf_frame)
    elif not isinstance(val, gdb.Value):
        val = gdb.Value(val)
    return (name, val)


class _ScopeReference(BaseReference):
    def __init__(self, name, hint, frameId: int, var_list):
        super().__init__(name)
        self.hint = hint
        self.frameId = frameId
        # VAR_LIST might be any kind of iterator, but it's convenient
        # here if it is just a collection.
        self.var_list = tuple(var_list)

    def to_object(self):
        result = super().to_object()
        result["presentationHint"] = self.hint
        # How would we know?
        result["expensive"] = False
        result["namedVariables"] = self.child_count()
        frame = frame_for_id(self.frameId)
        if frame.line() is not None:
            result["line"] = export_line(frame.line())
        filename = frame.filename()
        if filename is not None:
            result["source"] = make_source(filename)
        return result

    def has_children(self):
        return True

    def child_count(self):
        return len(self.var_list)

    @in_gdb_thread
    def fetch_one_child(self, idx):
        return symbol_value(self.var_list[idx], frame_for_id(self.frameId))


# A _ScopeReference that wraps the 'finish' value.  Note that this
# object is only created if such a value actually exists.
class _FinishScopeReference(_ScopeReference):
    def __init__(self, frameId):
        super().__init__("Return", "returnValue", frameId, ())

    def child_count(self):
        return 1

    def fetch_one_child(self, idx):
        assert idx == 0
        global _last_return_value
        return ("(return)", _last_return_value)


class _RegisterReference(_ScopeReference):
    def __init__(self, name, frameId):
        super().__init__(
            name,
            "registers",
            frameId,
            frame_for_id(frameId).inferior_frame().architecture().registers(),
        )

    @in_gdb_thread
    def fetch_one_child(self, idx):
        return (
            self.var_list[idx].name,
            frame_for_id(self.frameId)
            .inferior_frame()
            .read_register(self.var_list[idx]),
        )


@request("scopes")
def scopes(*, frameId: int, **extra):
    global _last_return_value
    global frame_to_scope
    if frameId in frame_to_scope:
        scopes = frame_to_scope[frameId]
    else:
        frame = frame_for_id(frameId)
        scopes = []
        # Make sure to handle the None case as well as the empty
        # iterator case.
        args = tuple(frame.frame_args() or ())
        if args:
            scopes.append(_ScopeReference("Arguments", "arguments", frameId, args))
        has_return_value = frameId == 0 and _last_return_value is not None
        # Make sure to handle the None case as well as the empty
        # iterator case.
        locs = tuple(frame.frame_locals() or ())
        if locs:
            scopes.append(_ScopeReference("Locals", "locals", frameId, locs))
        scopes.append(_RegisterReference("Registers", frameId))
        if has_return_value:
            scopes.append(_FinishScopeReference(frameId))
        frame_to_scope[frameId] = scopes
        global_scope = get_global_scope(frame)
        if global_scope is not None:
            scopes.append(global_scope)
    return {"scopes": [x.to_object() for x in scopes]}



# Copyright 2022-2024 Free Software Foundation, Inc.

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import itertools
from typing import Dict

import gdb
from gdb.frames import frame_iterator

from .startup import in_gdb_thread
from .state import set_thread

# A list of all the frames we've reported.  A frame's index in the
# list is its ID.  We don't use a hash here because frames are not
# hashable.
_all_frames = []


# Map from a global thread ID to a memoizing frame iterator.
_iter_map = {}

# Map from a global frame ID to a thread ID.
thread_ids: Dict[int, int] = {}


# Clear all the frame IDs.
@in_gdb_thread
def _clear_frame_ids(evt):
    global _all_frames
    _all_frames = []
    global _iter_map
    _iter_map = {}
    global thread_ids
    thread_ids = {}


# Clear the frame ID map whenever the inferior runs.
gdb.events.cont.connect(_clear_frame_ids)


@in_gdb_thread
def frame_for_id(id):
    """Given a frame identifier ID, return the corresponding frame."""
    global thread_ids
    if id in thread_ids:
        thread_id = thread_ids[id]
        if thread_id != gdb.selected_thread().global_num:
            set_thread(thread_id)
    global _all_frames
    return _all_frames[id]


@in_gdb_thread
def select_frame(id):
    """Given a frame identifier ID, select the corresponding frame."""
    frame = frame_for_id(id)
    frame.inferior_frame().select()


# A simple memoizing iterator.  Note that this is not very robust.
# For example, you can't start two copies of the iterator and then
# alternate fetching items from each.  Instead, it implements just
# what is needed for the current callers.
class _MemoizingIterator:
    def __init__(self, iterator):
        self.iterator = iterator
        self.seen = []

    def __iter__(self):
        # First the memoized items.
        for item in self.seen:
            yield item
        # Now memoize new items.
        for item in self.iterator:
            self.seen.append(item)
            yield item


# A generator that fetches frames and pairs them with a frame ID.  It
# yields tuples of the form (ID, ELIDED, FRAME), where ID is the
# generated ID, ELIDED is a boolean indicating if the frame should be
# elided, and FRAME is the frame itself.  This approach lets us
# memoize the result and assign consistent IDs, independent of how
# "includeAll" is set in the request.
@in_gdb_thread
def _frame_id_generator():
    try:
        base_iterator = frame_iterator(gdb.newest_frame(), 0, -1)
    except gdb.error:
        base_iterator = ()

    # Helper function to assign an ID to a frame.
    def get_id(frame):
        global _all_frames
        num = len(_all_frames)
        _all_frames.append(frame)
        global thread_ids
        thread_ids[num] = gdb.selected_thread().global_num
        return num

    def yield_frames(iterator, for_elided):
        for frame in iterator:
            # Unfortunately the frame filter docs don't describe
            # whether the elided frames conceptually come before or
            # after the eliding frame.  Here we choose after.
            yield (get_id(frame), for_elided, frame)

            elided = frame.elided()
            if elided is not None:
                yield from yield_frames(frame.elided(), True)

    yield from yield_frames(base_iterator, False)


# Return the memoizing frame iterator for the selected thread.
@in_gdb_thread
def _get_frame_iterator():
    thread_id = gdb.selected_thread().global_num
    global _iter_map
    if thread_id not in _iter_map:
        _iter_map[thread_id] = _MemoizingIterator(_frame_id_generator())
    return _iter_map[thread_id]


# A helper function that creates an iterable that returns (ID, FRAME)
# pairs.  It uses the memoizing frame iterator, but also handles the
# "includeAll" member of StackFrameFormat.
@in_gdb_thread
def dap_frame_generator(frame_low, levels, include_all):
    """A generator that yields identifiers and frames.

    Each element is a pair of the form (ID, FRAME).
    ID is the internally-assigned frame ID.
    FRAME is a FrameDecorator of some kind.

    Arguments are as to the stackTrace request."""

    base_iterator = _get_frame_iterator()

    if not include_all:
        base_iterator = itertools.filterfalse(lambda item: item[1], base_iterator)

    if levels == 0:
        # Zero means all remaining frames.
        frame_high = None
    else:
        frame_high = frame_low + levels
    base_iterator = itertools.islice(base_iterator, frame_low, frame_high)

    for ident, _, frame in base_iterator:
        yield (ident, frame)



# Copyright (C) 2021-2024 Free Software Foundation, Inc.

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Disassembler related module."""

import _gdb.disassembler

# Re-export everything from the _gdb.disassembler module, which is
# defined within GDB's C++ code.  Note that two indicators are needed
# here to silence flake8.
from _gdb.disassembler import *  # noqa: F401,F403

import gdb

# Module global dictionary of gdb.disassembler.Disassembler objects.
# The keys of this dictionary are bfd architecture names, or the
# special value None.
#
# When a request to disassemble comes in we first lookup the bfd
# architecture name from the gdbarch, if that name exists in this
# dictionary then we use that Disassembler object.
#
# If there's no architecture specific disassembler then we look for
# the key None in this dictionary, and if that key exists, we use that
# disassembler.
#
# If none of the above checks found a suitable disassembler, then no
# disassembly is performed in Python.
_disassemblers_dict = {}


class Disassembler(object):
    """A base class from which all user implemented disassemblers must
    inherit."""

    def __init__(self, name):
        """Constructor.  Takes a name, which should be a string, which can be
        used to identify this disassembler in diagnostic messages."""
        self.name = name

    def __call__(self, info):
        """A default implementation of __call__.  All sub-classes must
        override this method.  Calling this default implementation will throw
        a NotImplementedError exception."""
        raise NotImplementedError("Disassembler.__call__")


def register_disassembler(disassembler, architecture=None):
    """Register a disassembler.  DISASSEMBLER is a sub-class of
    gdb.disassembler.Disassembler.  ARCHITECTURE is either None or a
    string, the name of an architecture known to GDB.

    DISASSEMBLER is registered as a disassembler for ARCHITECTURE, or
    all architectures when ARCHITECTURE is None.

    Returns the previous disassembler registered with this
    ARCHITECTURE value.
    """

    if not isinstance(disassembler, Disassembler) and disassembler is not None:
        raise TypeError("disassembler should sub-class gdb.disassembler.Disassembler")

    old = None
    if architecture in _disassemblers_dict:
        old = _disassemblers_dict[architecture]
        del _disassemblers_dict[architecture]
    if disassembler is not None:
        _disassemblers_dict[architecture] = disassembler

    # Call the private _set_enabled function within the
    # _gdb.disassembler module.  This function sets a global flag
    # within GDB's C++ code that enables or dissables the Python
    # disassembler functionality, this improves performance of the
    # disassembler by avoiding unneeded calls into Python when we know
    # that no disassemblers are registered.
    _gdb.disassembler._set_enabled(len(_disassemblers_dict) > 0)
    return old


def _print_insn(info):
    """This function is called by GDB when it wants to disassemble an
    instruction.  INFO describes the instruction to be
    disassembled."""

    def lookup_disassembler(arch):
        name = arch.name()
        if name is None:
            return None
        if name in _disassemblers_dict:
            return _disassemblers_dict[name]
        if None in _disassemblers_dict:
            return _disassemblers_dict[None]
        return None

    disassembler = lookup_disassembler(info.architecture)
    if disassembler is None:
        return None
    return disassembler(info)


class maint_info_py_disassemblers_cmd(gdb.Command):
    """
    List all registered Python disassemblers.

    List the name of all registered Python disassemblers, next to the
    name of the architecture for which the disassembler is registered.

    The global Python disassembler is listed next to the string
    'GLOBAL'.

    The disassembler that matches the architecture of the currently
    selected inferior will be marked, this is an indication of which
    disassembler will be invoked if any disassembly is performed in
    the current inferior.
    """

    def __init__(self):
        super().__init__("maintenance info python-disassemblers", gdb.COMMAND_USER)

    def invoke(self, args, from_tty):
        # If no disassemblers are registered, tell the user.
        if len(_disassemblers_dict) == 0:
            print("No Python disassemblers registered.")
            return

        # Figure out the longest architecture name, so we can
        # correctly format the table of results.
        longest_arch_name = 0
        for architecture in _disassemblers_dict:
            if architecture is not None:
                name = _disassemblers_dict[architecture].name
                if len(name) > longest_arch_name:
                    longest_arch_name = len(name)

        # Figure out the name of the current architecture.  There
        # should always be a current inferior, but if, somehow, there
        # isn't, then leave curr_arch as the empty string, which will
        # not then match against any architecture in the dictionary.
        curr_arch = ""
        if gdb.selected_inferior() is not None:
            curr_arch = gdb.selected_inferior().architecture().name()

        # Now print the dictionary of registered disassemblers out to
        # the user.
        match_tag = "\t(Matches current architecture)"
        fmt_len = max(longest_arch_name, len("Architecture"))
        format_string = "{:" + str(fmt_len) + "s} {:s}"
        print(format_string.format("Architecture", "Disassember Name"))
        for architecture in _disassemblers_dict:
            if architecture is not None:
                name = _disassemblers_dict[architecture].name
                if architecture == curr_arch:
                    name += match_tag
                    match_tag = ""
                print(format_string.format(architecture, name))
        if None in _disassemblers_dict:
            name = _disassemblers_dict[None].name + match_tag
            print(format_string.format("GLOBAL", name))


maint_info_py_disassemblers_cmd()



# Copyright (C) 2010-2024 Free Software Foundation, Inc.

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import signal
import sys
import threading
import traceback
from contextlib import contextmanager

# Python 3 moved "reload"
if sys.version_info >= (3, 4):
    from importlib import reload
else:
    from imp import reload

import _gdb

# Note that two indicators are needed here to silence flake8.
from _gdb import *  # noqa: F401,F403

# isort: split

# Historically, gdb.events was always available, so ensure it's
# still available without an explicit import.
import _gdbevents as events

sys.modules["gdb.events"] = events


class _GdbFile(object):
    # These two are needed in Python 3
    encoding = "UTF-8"
    errors = "strict"

    def __init__(self, stream):
        self.stream = stream

    def close(self):
        # Do nothing.
        return None

    def isatty(self):
        return False

    def writelines(self, iterable):
        for line in iterable:
            self.write(line)

    def flush(self):
        _gdb.flush(stream=self.stream)

    def write(self, s):
        _gdb.write(s, stream=self.stream)


sys.stdout = _GdbFile(_gdb.STDOUT)
sys.stderr = _GdbFile(_gdb.STDERR)

# Default prompt hook does nothing.
prompt_hook = None

# Ensure that sys.argv is set to something.
# We do not use PySys_SetArgvEx because it did not appear until 2.6.6.
sys.argv = [""]

# Initial pretty printers.
pretty_printers = []

# Initial type printers.
type_printers = []
# Initial xmethod matchers.
xmethods = []
# Initial frame filters.
frame_filters = {}
# Initial frame unwinders.
frame_unwinders = []
# The missing file handlers.  Each item is a tuple with the form
# (TYPE, HANDLER) where TYPE is a string either 'debug' or 'objfile'.
missing_file_handlers = []


def _execute_unwinders(pending_frame):
    """Internal function called from GDB to execute all unwinders.

    Runs each currently enabled unwinder until it finds the one that
    can unwind given frame.

    Arguments:
        pending_frame: gdb.PendingFrame instance.

    Returns:
        Tuple with:

          [0] gdb.UnwindInfo instance
          [1] Name of unwinder that claimed the frame (type `str`)

        or None, if no unwinder has claimed the frame.
    """
    for objfile in objfiles():
        for unwinder in objfile.frame_unwinders:
            if unwinder.enabled:
                unwind_info = unwinder(pending_frame)
                if unwind_info is not None:
                    return (unwind_info, unwinder.name)

    for unwinder in current_progspace().frame_unwinders:
        if unwinder.enabled:
            unwind_info = unwinder(pending_frame)
            if unwind_info is not None:
                return (unwind_info, unwinder.name)

    for unwinder in frame_unwinders:
        if unwinder.enabled:
            unwind_info = unwinder(pending_frame)
            if unwind_info is not None:
                return (unwind_info, unwinder.name)

    return None


# Convenience variable to GDB's python directory
PYTHONDIR = os.path.dirname(os.path.dirname(__file__))

# Auto-load all functions/commands.

# Packages to auto-load.

packages = ["function", "command", "printer"]

# pkgutil.iter_modules is not available prior to Python 2.6.  Instead,
# manually iterate the list, collating the Python files in each module
# path.  Construct the module name, and import.


def _auto_load_packages():
    for package in packages:
        location = os.path.join(os.path.dirname(__file__), package)
        if os.path.exists(location):
            py_files = filter(
                lambda x: x.endswith(".py") and x != "__init__.py", os.listdir(location)
            )

            for py_file in py_files:
                # Construct from foo.py, gdb.module.foo
                modname = "%s.%s.%s" % (__name__, package, py_file[:-3])
                try:
                    if modname in sys.modules:
                        # reload modules with duplicate names
                        reload(__import__(modname))
                    else:
                        __import__(modname)
                except Exception:
                    sys.stderr.write(traceback.format_exc() + "\n")


_auto_load_packages()


def GdbSetPythonDirectory(dir):
    """Update sys.path, reload gdb and auto-load packages."""
    global PYTHONDIR

    try:
        sys.path.remove(PYTHONDIR)
    except ValueError:
        pass
    sys.path.insert(0, dir)

    PYTHONDIR = dir

    # note that reload overwrites the gdb module without deleting existing
    # attributes
    reload(__import__(__name__))
    _auto_load_packages()


def current_progspace():
    "Return the current Progspace."
    return _gdb.selected_inferior().progspace


def objfiles():
    "Return a sequence of the current program space's objfiles."
    return current_progspace().objfiles()


def solib_name(addr):
    """solib_name (Long) -> String.\n\
Return the name of the shared library holding a given address, or None."""
    return current_progspace().solib_name(addr)


def block_for_pc(pc):
    "Return the block containing the given pc value, or None."
    return current_progspace().block_for_pc(pc)


def find_pc_line(pc):
    """find_pc_line (pc) -> Symtab_and_line.
    Return the gdb.Symtab_and_line object corresponding to the pc value."""
    return current_progspace().find_pc_line(pc)


def set_parameter(name, value):
    """Set the GDB parameter NAME to VALUE."""
    # Handle the specific cases of None and booleans here, because
    # gdb.parameter can return them, but they can't be passed to 'set'
    # this way.
    if value is None:
        value = "unlimited"
    elif isinstance(value, bool):
        if value:
            value = "on"
        else:
            value = "off"
    _gdb.execute("set " + name + " " + str(value), to_string=True)


@contextmanager
def with_parameter(name, value):
    """Temporarily set the GDB parameter NAME to VALUE.
    Note that this is a context manager."""
    old_value = _gdb.parameter(name)
    set_parameter(name, value)
    try:
        # Nothing that useful to return.
        yield None
    finally:
        set_parameter(name, old_value)


@contextmanager
def blocked_signals():
    """A helper function that blocks and unblocks signals."""
    if not hasattr(signal, "pthread_sigmask"):
        yield
        return

    to_block = {signal.SIGCHLD, signal.SIGINT, signal.SIGALRM, signal.SIGWINCH}
    old_mask = signal.pthread_sigmask(signal.SIG_BLOCK, to_block)
    try:
        yield None
    finally:
        signal.pthread_sigmask(signal.SIG_SETMASK, old_mask)


class Thread(threading.Thread):
    """A GDB-specific wrapper around threading.Thread

    This wrapper ensures that the new thread blocks any signals that
    must be delivered on GDB's main thread."""

    def start(self):
        # GDB requires that these be delivered to the main thread.  We
        # do this here to avoid any possible race with the creation of
        # the new thread.  The thread mask is inherited by new
        # threads.
        with blocked_signals():
            super().start()


def _filter_missing_file_handlers(handlers, handler_type):
    """Each list of missing file handlers is a list of tuples, the first
    item in the tuple is a string either 'debug' or 'objfile' to
    indicate what type of handler it is.  The second item in the tuple
    is the actual handler object.

    This function takes HANDLER_TYPE which is a string, either 'debug'
    or 'objfile' and HANDLERS, a list of tuples.  The function returns
    an iterable over all of the handler objects (extracted from the
    tuples) which match HANDLER_TYPE.
    """

    return map(lambda t: t[1], filter(lambda t: t[0] == handler_type, handlers))


def _handle_missing_files(pspace, handler_type, cb):
    """Helper for _handle_missing_debuginfo and _handle_missing_objfile.

    Arguments:
        pspace: The gdb.Progspace in which we're operating.  Used to
            lookup program space specific handlers.
        handler_type: A string, either 'debug' or 'objfile', this is the
            type of handler we're looking for.
        cb: A callback which takes a handler and returns the result of
            calling the handler.

    Returns:
        None: No suitable file could be found.
        False: A handler has decided that the requested file cannot be
                found, and no further searching should be done.
        True: The file has been found and installed in a location
                where GDB would normally look for it.  GDB should
                repeat its lookup process, the file should now be in
                place.
        A string: This is the filename of where the missing file can
                be found.
    """

    for handler in _filter_missing_file_handlers(
        pspace.missing_file_handlers, handler_type
    ):
        if handler.enabled:
            result = cb(handler)
            if result is not None:
                return result

    for handler in _filter_missing_file_handlers(missing_file_handlers, handler_type):
        if handler.enabled:
            result = cb(handler)
            if result is not None:
                return result

    return None


def _handle_missing_debuginfo(objfile):
    """Internal function called from GDB to execute missing debug
    handlers.

    Run each of the currently registered, and enabled missing debug
    handler objects for the current program space and then from the
    global list.  Stop after the first handler that returns a result
    other than None.

    Arguments:
        objfile: A gdb.Objfile for which GDB could not find any debug
                 information.

    Returns:
        None: No debug information could be found for objfile.
        False: A handler has done all it can with objfile, but no
               debug information could be found.
        True: Debug information might have been installed by a
              handler, GDB should check again.
        A string: This is the filename of a file containing the
                  required debug information.
    """

    pspace = objfile.progspace

    return _handle_missing_files(pspace, "debug", lambda h: h(objfile))


def _handle_missing_objfile(pspace, buildid, filename):
    """Internal function called from GDB to execute missing objfile
    handlers.

    Run each of the currently registered, and enabled missing objfile
    handler objects for the gdb.Progspace passed in as an argument,
    and then from the global list.  Stop after the first handler that
    returns a result other than None.

    Arguments:
        pspace: A gdb.Progspace for which the missing objfile handlers
                should be run.  This is the program space in which an
                objfile was found to be missing.
        buildid: A string containing the build-id we're looking for.
        filename: The filename of the file GDB tried to find but
                  couldn't.  This is not where the file should be
                  placed if found, in fact, this file might already
                  exist on disk but have the wrong build-id.  This is
                  mostly provided in order to be used in messages to
                  the user.

    Returns:
        None: No objfile could be found for this build-id.
        False: A handler has done all it can with for this build-id,
               but no objfile could be found.
        True: An objfile might have been installed by a handler, GDB
              should check again.  The only place GDB checks is within
              the .build-id sub-directory within the
              debug-file-directory.  If the required file was not
              installed there then GDB will not find it.
        A string: This is the filename of a file containing the
                  missing objfile.
    """

    return _handle_missing_files(
        pspace, "objfile", lambda h: h(pspace, buildid, filename)
    )



# Styling related hooks.
# Copyright (C) 2010-2024 Free Software Foundation, Inc.

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities for styling."""

import gdb

try:
    from pygments import formatters, highlight, lexers
    from pygments.filters import TokenMergeFilter
    from pygments.token import Comment, Error, Text

    _formatter = None

    def get_formatter():
        global _formatter
        if _formatter is None:
            _formatter = formatters.TerminalFormatter()
        return _formatter

    def colorize(filename, contents):
        # Don't want any errors.
        try:
            lexer = lexers.get_lexer_for_filename(filename, stripnl=False)
            formatter = get_formatter()
            return highlight(contents, lexer, formatter).encode(
                gdb.host_charset(), "backslashreplace"
            )
        except Exception:
            return None

    class HandleNasmComments(TokenMergeFilter):
        @staticmethod
        def fix_comments(lexer, stream):
            in_comment = False
            for ttype, value in stream:
                if ttype is Error and value == "#":
                    in_comment = True
                if in_comment:
                    if ttype is Text and value == "\n":
                        in_comment = False
                    else:
                        ttype = Comment.Single
                yield ttype, value

        def filter(self, lexer, stream):
            f = HandleNasmComments.fix_comments
            return super().filter(lexer, f(lexer, stream))

    _asm_lexers = {}

    def __get_asm_lexer(gdbarch):
        lexer_type = "asm"
        try:
            # For an i386 based architecture, in 'intel' mode, use the nasm
            # lexer.
            flavor = gdb.parameter("disassembly-flavor")
            if flavor == "intel" and gdbarch.name()[:4] == "i386":
                lexer_type = "nasm"
        except Exception:
            # If GDB is built without i386 support then attempting to fetch
            # the 'disassembly-flavor' parameter will throw an error, which we
            # ignore.
            pass

        global _asm_lexers
        if lexer_type not in _asm_lexers:
            _asm_lexers[lexer_type] = lexers.get_lexer_by_name(lexer_type)
            _asm_lexers[lexer_type].add_filter(HandleNasmComments())
            _asm_lexers[lexer_type].add_filter("raiseonerror")
        return _asm_lexers[lexer_type]

    def colorize_disasm(content, gdbarch):
        # Don't want any errors.
        try:
            lexer = __get_asm_lexer(gdbarch)
            formatter = get_formatter()
            return highlight(content, lexer, formatter).rstrip().encode()
        except Exception:
            return content

except ImportError:

    def colorize(filename, contents):
        return None

    def colorize_disasm(content, gdbarch):
        return None



# Extended prompt utilities.
# Copyright (C) 2011-2024 Free Software Foundation, Inc.

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" Extended prompt library functions."""

import os

import gdb


def _prompt_pwd(ignore):
    "The current working directory."
    return os.getcwd()


def _prompt_object_attr(func, what, attr, nattr):
    """Internal worker for fetching GDB attributes."""
    if attr is None:
        attr = nattr
    try:
        obj = func()
    except gdb.error:
        return "<no %s>" % what
    if hasattr(obj, attr):
        result = getattr(obj, attr)
        if callable(result):
            result = result()
        return result
    else:
        return "<no attribute %s on current %s>" % (attr, what)


def _prompt_frame(attr):
    "The selected frame; an argument names a frame parameter."
    return _prompt_object_attr(gdb.selected_frame, "frame", attr, "name")


def _prompt_thread(attr):
    "The selected thread; an argument names a thread parameter."
    return _prompt_object_attr(gdb.selected_thread, "thread", attr, "num")


def _prompt_version(attr):
    "The version of GDB."
    return gdb.VERSION


def _prompt_esc(attr):
    "The ESC character."
    return "\033"


def _prompt_bs(attr):
    "A backslash."
    return "\\"


def _prompt_n(attr):
    "A newline."
    return "\n"


def _prompt_r(attr):
    "A carriage return."
    return "\r"


def _prompt_param(attr):
    "A parameter's value; the argument names the parameter."
    return gdb.parameter(attr)


def _prompt_noprint_begin(attr):
    "Begins a sequence of non-printing characters."
    return "\001"


def _prompt_noprint_end(attr):
    "Ends a sequence of non-printing characters."
    return "\002"


prompt_substitutions = {
    "e": _prompt_esc,
    "\\": _prompt_bs,
    "n": _prompt_n,
    "r": _prompt_r,
    "v": _prompt_version,
    "w": _prompt_pwd,
    "f": _prompt_frame,
    "t": _prompt_thread,
    "p": _prompt_param,
    "[": _prompt_noprint_begin,
    "]": _prompt_noprint_end,
}


def prompt_help():
    """Generate help dynamically from the __doc__ strings of attribute
    functions."""

    result = ""
    keys = sorted(prompt_substitutions.keys())
    for key in keys:
        result += "  \\%s\t%s\n" % (key, prompt_substitutions[key].__doc__)
    result += """
A substitution can be used in a simple form, like "\\f".
An argument can also be passed to it, like "\\f{name}".
The meaning of the argument depends on the particular substitution."""
    return result


def substitute_prompt(prompt):
    "Perform substitutions on PROMPT."

    result = ""
    plen = len(prompt)
    i = 0
    while i < plen:
        if prompt[i] == "\\":
            i = i + 1
            if i >= plen:
                break
            cmdch = prompt[i]

            if cmdch in prompt_substitutions:
                cmd = prompt_substitutions[cmdch]

                if i + 1 < plen and prompt[i + 1] == "{":
                    j = i + 1
                    while j < plen and prompt[j] != "}":
                        j = j + 1
                    # Just ignore formatting errors.
                    if j >= plen or prompt[j] != "}":
                        arg = None
                    else:
                        arg = prompt[i + 2 : j]
                        i = j
                else:
                    arg = None
                result += str(cmd(arg))
            else:
                # Unrecognized escapes are turned into the escaped
                # character itself.
                result += prompt[i]
        else:
            result += prompt[i]

        i = i + 1

    return result



# Pretty-printers for ROCm OCP microscaling floating point types
# Copyright (C) 2025 Free Software Foundation, Inc.

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import gdb
import gdb.printing
from gdb.function.rocm_get_scaled_float import (
    compute_uint8e8m0,
    extract_mx_values,
    hipexp_ocp_fp_re,
)


class OcpMxFpValuePrinter(gdb.ValuePrinter):
    def __init__(self, val):
        self._val = val

    def to_string(self):
        fp = extract_mx_values(self._val)
        return "{{{}}}".format(", ".join([str(f) for f in fp]))


class OcpMxScaleValuePrinter(gdb.ValuePrinter):
    def __init__(self, val):
        self._val = val

    def to_string(self):
        return compute_uint8e8m0(self._val)


def build_ocp_mx_subprinters():
    pp = gdb.printing.RegexpCollectionPrettyPrinter("rocm_ocp_mx")
    pp.add_printer("hipext_ocp_fp", hipexp_ocp_fp_re(), OcpMxFpValuePrinter)
    pp.add_printer("amd_scale_t", "^__amd_scale_t$", OcpMxScaleValuePrinter)
    return pp


# Register pretty printer collection automatically when the script is sourced by GDB
gdb.printing.register_pretty_printer(gdb.current_objfile(), build_ocp_mx_subprinters())



# Copyright 2022-2024 Free Software Foundation, Inc.

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# This is deprecated in 3.9, but required in older versions.
from typing import Optional

import gdb

from .frames import select_frame
from .server import capability, client_bool_capability, request
from .startup import DAPException, in_gdb_thread, parse_and_eval
from .varref import VariableReference, apply_format, find_variable


class EvaluateResult(VariableReference):
    def __init__(self, value):
        super().__init__(None, value, "result")


# Helper function to evaluate an expression in a certain frame.
@in_gdb_thread
def _evaluate(expr, frame_id, value_format):
    with apply_format(value_format):
        global_context = True
        if frame_id is not None:
            select_frame(frame_id)
            global_context = False
        val = parse_and_eval(expr, global_context=global_context)
        ref = EvaluateResult(val)
        return ref.to_object()


# Like _evaluate but ensure that the expression cannot cause side
# effects.
@in_gdb_thread
def _eval_for_hover(expr, frame_id, value_format):
    with gdb.with_parameter("may-write-registers", "off"):
        with gdb.with_parameter("may-write-memory", "off"):
            with gdb.with_parameter("may-call-functions", "off"):
                return _evaluate(expr, frame_id, value_format)


class _SetResult(VariableReference):
    def __init__(self, value):
        super().__init__(None, value, "value")


# Helper function to evaluate a gdb command in a certain frame.
@in_gdb_thread
def _repl(command, frame_id):
    if frame_id is not None:
        select_frame(frame_id)
    val = gdb.execute(command, from_tty=True, to_string=True)
    return {
        "result": val,
        "variablesReference": 0,
    }


@request("evaluate")
@capability("supportsEvaluateForHovers")
@capability("supportsValueFormattingOptions")
def eval_request(
    *,
    expression: str,
    frameId: Optional[int] = None,
    context: str = "variables",
    format=None,
    **args,
):
    if context in ("watch", "variables"):
        # These seem to be expression-like.
        return _evaluate(expression, frameId, format)
    elif context == "hover":
        return _eval_for_hover(expression, frameId, format)
    elif context == "repl":
        # Ignore the format for repl evaluation.
        return _repl(expression, frameId)
    else:
        raise DAPException('unknown evaluate context "' + context + '"')


@request("variables")
# Note that we ignore the 'filter' field.  That seems to be
# specific to javascript.
def variables(
    *, variablesReference: int, start: int = 0, count: int = 0, format=None, **args
):
    # This behavior was clarified here:
    # https://github.com/microsoft/debug-adapter-protocol/pull/394
    if not client_bool_capability("supportsVariablePaging"):
        start = 0
        count = 0
    with apply_format(format):
        var = find_variable(variablesReference)
        children = var.fetch_children(start, count)
        return {"variables": [x.to_object() for x in children]}


@capability("supportsSetExpression")
@request("setExpression")
def set_expression(
    *, expression: str, value: str, frameId: Optional[int] = None, format=None, **args
):
    with apply_format(format):
        global_context = True
        if frameId is not None:
            select_frame(frameId)
            global_context = False
        lhs = parse_and_eval(expression, global_context=global_context)
        rhs = parse_and_eval(value, global_context=global_context)
        lhs.assign(rhs)
        return _SetResult(lhs).to_object()


@capability("supportsSetVariable")
@request("setVariable")
def set_variable(
    *, variablesReference: int, name: str, value: str, format=None, **args
):
    with apply_format(format):
        var = find_variable(variablesReference)
        lhs = var.find_child_by_name(name)
        rhs = parse_and_eval(value)
        lhs.assign(rhs)
        return lhs.to_object()



# Copyright (C) 2015-2024 Free Software Foundation, Inc.

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Unwinder class and register_unwinder function."""

import gdb


class Unwinder(object):
    """Base class (or a template) for frame unwinders written in Python.

    An unwinder has a single method __call__ and the attributes
    described below.

    Attributes:
        name: The name of the unwinder.
        enabled: A boolean indicating whether the unwinder is enabled.
    """

    def __init__(self, name):
        """Constructor.

        Args:
            name: An identifying name for the unwinder.
        """

        if not isinstance(name, str):
            raise TypeError("incorrect type for name: %s" % type(name))

        self._name = name
        self._enabled = True

    @property
    def name(self):
        return self._name

    @property
    def enabled(self):
        return self._enabled

    @enabled.setter
    def enabled(self, value):
        if not isinstance(value, bool):
            raise TypeError("incorrect type for enabled attribute: %s" % type(value))
        self._enabled = value
        gdb.invalidate_cached_frames()

    def __call__(self, pending_frame):
        """GDB calls this method to unwind a frame.

        Arguments:
            pending_frame: gdb.PendingFrame instance.

        Returns:
            gdb.UnwindInfo instance.
        """
        raise NotImplementedError("Unwinder __call__.")


class FrameId(object):
    """A Frame-ID class for use when creating gdb.UnwindInfo objects.

    Attributes (all read-only):
        pc: Program counter value.
        sp: The stack-pointer value.
        special: An alternative stack-pointer value, can be None."""

    def __init__(self, sp, pc, special=None):
        self._sp = sp
        self._pc = pc
        self._special = special

    @property
    def sp(self):
        return self._sp

    @property
    def pc(self):
        return self._pc

    @property
    def special(self):
        return self._special


def register_unwinder(locus, unwinder, replace=False):
    """Register unwinder in given locus.

    The unwinder is prepended to the locus's unwinders list. Unwinder
    name should be unique.

    Arguments:
        locus: Either an objfile, progspace, or None (in which case
               the unwinder is registered globally).
        unwinder: An object of a gdb.Unwinder subclass
        replace: If True, replaces existing unwinder with the same name.
                 Otherwise, raises exception if unwinder with the same
                 name already exists.

    Returns:
        Nothing.

    Raises:
        RuntimeError: Unwinder name is not unique
        TypeError: Bad locus type
    """
    if locus is None:
        if gdb.parameter("verbose"):
            gdb.write("Registering global %s unwinder ...\n" % unwinder.name)
        locus = gdb
    elif isinstance(locus, gdb.Objfile) or isinstance(locus, gdb.Progspace):
        if gdb.parameter("verbose"):
            gdb.write(
                "Registering %s unwinder for %s ...\n" % (unwinder.name, locus.filename)
            )
    else:
        raise TypeError("locus should be gdb.Objfile or gdb.Progspace or None")

    i = 0
    for needle in locus.frame_unwinders:
        if needle.name == unwinder.name:
            if replace:
                del locus.frame_unwinders[i]
            else:
                raise RuntimeError("Unwinder %s already exists." % unwinder.name)
        i += 1
    locus.frame_unwinders.insert(0, unwinder)
    gdb.invalidate_cached_frames()



# Copyright (C) 2011-2024 Free Software Foundation, Inc.

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Configure GDB using the ELinOS environment."""

import glob
import os

import gdb


def warn(msg):
    print("warning: %s" % msg)


def get_elinos_environment():
    """Return the ELinOS environment.

    If the ELinOS environment is properly set up, return a dictionary
    which contains:
      * The path to the ELinOS project at key 'project';
      * The path to the ELinOS CDK at key 'cdk';
      * The ELinOS target name at key 'target' (Eg. 'i486-linux');
      * A list of Xenomai install prefixes (which could be empty, if
        the ELinOS project does not include Xenomai) at key 'xenomai'.

    If one of these cannot be found, print a warning; the corresponding
    value in the returned dictionary will be None.
    """
    result = {}
    for key in ("project", "cdk", "target"):
        var = "ELINOS_" + key.upper()
        if var in os.environ:
            result[key] = os.environ[var]
        else:
            warn("%s not set" % var)
            result[key] = None

    if result["project"] is not None:
        result["xenomai"] = glob.glob(result["project"] + "/xenomai-[0-9.]*")
    else:
        result["xenomai"] = []

    return result


def elinos_init():
    """Initialize debugger environment for ELinOS.

    Let the debugger know where to find the ELinOS libraries on host. This
    assumes that an ELinOS environment is properly set up. If some environment
    variables are missing, warn about which library may be missing.
    """
    elinos_env = get_elinos_environment()

    solib_dirs = []

    # System libraries
    if None in (elinos_env[key] for key in ("cdk", "target")):
        warn("ELinOS system libraries will not be loaded")
    else:
        solib_prefix = "%s/%s" % (elinos_env["cdk"], elinos_env["target"])
        solib_dirs += ["%s/%s" % (solib_prefix, "lib")]
        gdb.execute("set solib-absolute-prefix %s" % solib_prefix)

    # Xenomai libraries. Those are optional, so have a lighter warning
    # if they cannot be located.
    if elinos_env["project"] is None:
        warn("Xenomai libraries may not be loaded")
    else:
        for dir in elinos_env["xenomai"]:
            solib_dirs += ["%s/%s" % (dir, "xenomai-build/usr/realtime/lib")]

    if len(solib_dirs) != 0:
        gdb.execute("set solib-search-path %s" % ":".join(solib_dirs))


if __name__ == "__main__":
    elinos_init()



# Missing debug and objfile related commands.
#
# Copyright 2023-2024 Free Software Foundation, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re

import gdb


def validate_regexp(exp, idstring):
    """Compile exp into a compiled regular expression object.

    Arguments:
        exp: The string to compile into a re.Pattern object.
        idstring: A string, what exp is a regexp for.

    Returns:
        A re.Pattern object representing exp.

    Raises:
        SyntaxError: If exp is an invalid regexp.
    """

    try:
        return re.compile(exp)
    except SyntaxError:
        raise SyntaxError("Invalid %s regexp: %s." % (idstring, exp))


def parse_missing_file_command_args(arg):
    """Internal utility to parse missing file handler command argv.

    Arguments:
        arg: The arguments to the command. The format is:
             [locus-regexp [name-regexp]]

    Returns:
        A 2-tuple of compiled regular expressions.

    Raises:
        SyntaxError: an error processing ARG
    """

    argv = gdb.string_to_argv(arg)
    argc = len(argv)
    if argc > 2:
        raise SyntaxError("Too many arguments.")
    locus_regexp = ""
    name_regexp = ""
    if argc >= 1:
        locus_regexp = argv[0]
        if argc >= 2:
            name_regexp = argv[1]
    return (
        validate_regexp(locus_regexp, "locus"),
        validate_regexp(name_regexp, "handler"),
    )


class InfoMissingFileHandlers(gdb.Command):
    """GDB command to list missing HTYPE handlers.

    Usage: info missing-HTYPE-handlers [LOCUS-REGEXP [NAME-REGEXP]]

    LOCUS-REGEXP is a regular expression matching the location of the
    handler.  If it is omitted, all registered handlers from all
    loci are listed.  A locus can be 'global', 'progspace' to list
    the handlers from the current progspace, or a regular expression
    matching filenames of progspaces.

    NAME-REGEXP is a regular expression to filter missing HTYPE
    handler names.  If this omitted for a specified locus, then all
    registered handlers in the locus are listed.
    """

    def __init__(self, handler_type):
        # Update the doc string before calling the parent constructor,
        # replacing the string 'HTYPE' with the value of HANDLER_TYPE.
        # The parent constructor will grab a copy of this string to
        # use as the commands help text.
        self.__doc__ = self.__doc__.replace("HTYPE", handler_type)
        super().__init__(
            "info missing-" + handler_type + "-handlers", gdb.COMMAND_FILES
        )
        self.handler_type = handler_type

    def list_handlers(self, title, handlers, name_re):
        """Lists the missing file handlers whose name matches regexp.

        Arguments:
            title: The line to print before the list.
            handlers: The list of the missing file handlers.
            name_re: handler name filter.
        """

        if not handlers:
            return
        print(title)
        for handler in gdb._filter_missing_file_handlers(handlers, self.handler_type):
            if name_re.match(handler.name):
                print(
                    "  %s%s" % (handler.name, "" if handler.enabled else " [disabled]")
                )

    def invoke(self, arg, from_tty):
        locus_re, name_re = parse_missing_file_command_args(arg)

        if locus_re.match("progspace") and locus_re.pattern != "":
            cp = gdb.current_progspace()
            self.list_handlers(
                "Progspace %s:" % cp.filename, cp.missing_file_handlers, name_re
            )

        for progspace in gdb.progspaces():
            filename = progspace.filename or ""
            if locus_re.match(filename):
                if filename == "":
                    if progspace == gdb.current_progspace():
                        msg = "Current Progspace:"
                    else:
                        msg = "Progspace <no-file>:"
                else:
                    msg = "Progspace %s:" % filename
                self.list_handlers(
                    msg,
                    progspace.missing_file_handlers,
                    name_re,
                )

        # Print global handlers last, as these are invoked last.
        if locus_re.match("global"):
            self.list_handlers("Global:", gdb.missing_file_handlers, name_re)


def do_enable_handler1(handlers, name_re, flag, handler_type):
    """Enable/disable missing file handlers whose names match given regex.

    Arguments:
        handlers: The list of missing file handlers.
        name_re: Handler name filter.
        flag: A boolean indicating if we should enable or disable.
        handler_type: A string, either 'debug' or 'objfile', use to control
            which handlers are modified.

    Returns:
        The number of handlers affected.
    """

    total = 0
    for handler in gdb._filter_missing_file_handlers(handlers, handler_type):
        if name_re.match(handler.name) and handler.enabled != flag:
            handler.enabled = flag
            total += 1
    return total


def do_enable_handler(arg, flag, handler_type):
    """Enable or disable missing file handlers."""

    (locus_re, name_re) = parse_missing_file_command_args(arg)
    total = 0
    if locus_re.match("global"):
        total += do_enable_handler1(
            gdb.missing_file_handlers, name_re, flag, handler_type
        )
    if locus_re.match("progspace") and locus_re.pattern != "":
        total += do_enable_handler1(
            gdb.current_progspace().missing_file_handlers, name_re, flag, handler_type
        )
    for progspace in gdb.progspaces():
        filename = progspace.filename or ""
        if locus_re.match(filename):
            total += do_enable_handler1(
                progspace.missing_file_handlers, name_re, flag, handler_type
            )
    print(
        "%d missing %s handler%s %s"
        % (
            total,
            handler_type,
            "" if total == 1 else "s",
            "enabled" if flag else "disabled",
        )
    )


class EnableMissingFileHandler(gdb.Command):
    """GDB command to enable missing HTYPE handlers.

    Usage: enable missing-HTYPE-handler [LOCUS-REGEXP [NAME-REGEXP]]

    LOCUS-REGEXP is a regular expression specifying the handlers to
    enable.  It can be 'global', 'progspace' for the current
    progspace, or the filename for a file associated with a progspace.

    NAME_REGEXP is a regular expression to filter handler names.  If
    this omitted for a specified locus, then all registered handlers
    in the locus are affected.
    """

    def __init__(self, handler_type):
        # Update the doc string before calling the parent constructor,
        # replacing the string 'HTYPE' with the value of HANDLER_TYPE.
        # The parent constructor will grab a copy of this string to
        # use as the commands help text.
        self.__doc__ = self.__doc__.replace("HTYPE", handler_type)
        super().__init__(
            "enable missing-" + handler_type + "-handler", gdb.COMMAND_FILES
        )
        self.handler_type = handler_type

    def invoke(self, arg, from_tty):
        """GDB calls this to perform the command."""
        do_enable_handler(arg, True, self.handler_type)


class DisableMissingFileHandler(gdb.Command):
    """GDB command to disable missing HTYPE handlers.

    Usage: disable missing-HTYPE-handler [LOCUS-REGEXP [NAME-REGEXP]]

    LOCUS-REGEXP is a regular expression specifying the handlers to
    enable.  It can be 'global', 'progspace' for the current
    progspace, or the filename for a file associated with a progspace.

    NAME_REGEXP is a regular expression to filter handler names.  If
    this omitted for a specified locus, then all registered handlers
    in the locus are affected.
    """

    def __init__(self, handler_type):
        # Update the doc string before calling the parent constructor,
        # replacing the string 'HTYPE' with the value of HANDLER_TYPE.
        # The parent constructor will grab a copy of this string to
        # use as the commands help text.
        self.__doc__ = self.__doc__.replace("HTYPE", handler_type)
        super().__init__(
            "disable missing-" + handler_type + "-handler", gdb.COMMAND_FILES
        )
        self.handler_type = handler_type

    def invoke(self, arg, from_tty):
        """GDB calls this to perform the command."""
        do_enable_handler(arg, False, self.handler_type)


def register_missing_file_handler_commands():
    """Installs the missing file handler commands."""
    for handler_type in ["debug", "objfile"]:
        InfoMissingFileHandlers(handler_type)
        EnableMissingFileHandler(handler_type)
        DisableMissingFileHandler(handler_type)


register_missing_file_handler_commands()



# Copyright 2022-2024 Free Software Foundation, Inc.

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import gdb

from .events import exec_and_expect_stop
from .server import capability, request, send_gdb, send_gdb_with_response
from .startup import in_gdb_thread
from .state import set_thread


# Helper function to set the current thread and the scheduler-locking
# mode.  Returns True if scheduler-locking was successfully set to
# 'on', False in all other cases, including error.  When SELECT is
# True, also select that thread's newest frame.
@in_gdb_thread
def _handle_thread_step(thread_id, single_thread, select=False):
    # Ensure we're going to step the correct thread.
    set_thread(thread_id)
    if single_thread:
        result = True
        arg = "on"
    else:
        result = False
        arg = "off"
    try:
        # This can fail, depending on the target, so catch the error
        # and report to our caller.  We can't use exec_and_log because
        # that does not propagate exceptions.
        gdb.execute("set scheduler-locking " + arg, from_tty=True, to_string=True)
    except gdb.error:
        result = False
    # Other DAP code may select a frame, and the "finish" command uses
    # the selected frame.
    if select:
        gdb.newest_frame().select()
    return result


@request("next", response=False)
def next(
    *, threadId: int, singleThread: bool = False, granularity: str = "statement", **args
):
    _handle_thread_step(threadId, singleThread)
    cmd = "next"
    if granularity == "instruction":
        cmd += "i"
    exec_and_expect_stop(cmd)


@capability("supportsSteppingGranularity")
@capability("supportsSingleThreadExecutionRequests")
@request("stepIn", response=False)
def step_in(
    *, threadId: int, singleThread: bool = False, granularity: str = "statement", **args
):
    _handle_thread_step(threadId, singleThread)
    cmd = "step"
    if granularity == "instruction":
        cmd += "i"
    exec_and_expect_stop(cmd)


@request("stepOut", defer_stop_events=True)
def step_out(*, threadId: int, singleThread: bool = False, **args):
    _handle_thread_step(threadId, singleThread, True)
    exec_and_expect_stop("finish &", propagate_exception=True)


# This is a server-side request because it is funny: it wants to
# 'continue' but also return a result, which precludes using
# response=False.  Using 'continue &' would mostly work ok, but this
# yields races when a stop occurs before the response is sent back to
# the client.
@request("continue", on_dap_thread=True)
def continue_request(*, threadId: int, singleThread: bool = False, **args):
    locked = send_gdb_with_response(lambda: _handle_thread_step(threadId, singleThread))
    send_gdb(lambda: exec_and_expect_stop("continue"))
    return {"allThreadsContinued": not locked}
